<'
# Copyright (c) 2012-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Mon Apr  9 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#ifdef __linux__
# if defined(HAVE_VASPRINTF) && defined(HAVE_ASPRINTF)
/* Apparently we need _GNU_SOURCE defined to get access to vasprintf on Linux */
#  ifndef _GNU_SOURCE
#   define _GNU_SOURCE
#  endif
# endif
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "<"$comp">/c/client.h"
#include "<"$comp">_json-common.h"
#include "<"$comp">_typecopy.h"

<'foreach s [$component services] {'>

/* <"[--- Service [$s name] encoding/decoding ---------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_json_scan(
  struct genom_<"$comp">_<"[$s name]">_input *in, const char *json,
  char **endptr)
{
  int s = 0;
  json_skip_whitespace(json);
  if (*(json++) != '{') { s = EINVAL; goto done; }
  json_skip_whitespace(json);
  while(*json != '}') {
    if (*(json++) != '"') { s = EINVAL; goto done; };
<'    foreach p [$s parameters in inout] {'>
    if (!strncmp(json, "<"[$p name]">\"", 1+<"[string length [$p name]]">)) {
      json += 1+<"[string length [$p name]]">;
      json_skip_whitespace(json);
      if (*(json++) != ':') { s = EINVAL; goto done; }
      if ((s = json_scan_<"[[$p type] mangle]">(
             <"[[$p type] pass reference in->[$p name]]">, &json)))
        goto done;
    } else
<'    }'>
    { s = EINVAL; goto done; }

    json_skip_whitespace(json);
    if (*json == ',') {
      json++;
      json_skip_whitespace(json);
    }
  }

done:
  if (endptr) *endptr = (char *)json;
  if (s) {
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
  return genom_ok;
}

genom_event
genom_<"$comp">_client_<"[$s name]">_json_print(char **json,
             const struct genom_<"$comp">_<"[$s name]">_output *out)
{
  int s = 0;
  *json = NULL;

  if ((s = bufcat(json, "{"))) {
    if (*json) free(*json);
    *json = NULL;
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

<'  set sep {}'>
<'  foreach p [$s parameters inout out] {'>
  if ((s = bufcat(json, "<"$sep">\"<"[$p name]">\":"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
  if ((s = json_print_<"[[$p type] mangle]">(json,
         <"[[$p type] pass value out->[$p name]]">))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
<'  set sep ,'>
<'    }'>

  if ((s = bufcat(json, "}"))) {
    if (*json) free(*json);
    *json = NULL;
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  return genom_ok;
}

<'}'>

<'foreach p [$component port out] {'>

/* <"[--- Port [$p name] encoding ---------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_json_print(char **json,
  <"[[$p datatype] argument value data]">)
{
  int s;
  *json = NULL;

  if ((s = bufcat(json, "{"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  if ((s = bufcat(json, "\"<"[$p name]">\":"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
  if ((s = json_print_<"[[$p datatype] mangle]">(json, data))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  if ((s = bufcat(json, "}"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  return genom_ok;
}

<'}'>


/* <"[--- ${comp}_json_error --------------------------------------------]"> */

char *
genom_<"$comp">_client_json_error(genom_event e, const void *detail)
{
  char *s = NULL;

  if (!e) detail = genom_thrown(&e);

  if (e == genom_syserr_id) {
    if (detail) {
      uint32_t e = ((genom_syserr_detail *)detail)->code;
      if (bufcat(&s, "\"%s\"", strerror(e))) return NULL;
    } else {
      if (bufcat(&s, "\"%s\"", genom_syserr_id)) return NULL;
    }
    return s;
  } else if (e == genom_unkex_id && detail) {
    if (detail) {
      if (bufcat(&s, "{\"ex\":\"%s\",\"detail\":{\"what\":\"%s\"}}",
                 genom_unkex_id, ((genom_unkex_detail *)detail)->what))
        return NULL;
    } else {
      if (bufcat(&s, "{\"ex\":\"%s\"}", genom_unkex_id)) return NULL;
    }
    return s;
  }
<'foreach e [$component throws] {'>
  else if (e == <"[$e cname]">_id) {
<'  if {[llength [$e members]]} {'>
    if (detail) {
      if (bufcat(&s, "{\"ex\":\"%s\",\"detail\":", <"[$e cname]">_id))
        return NULL;
      if (json_print_<"[$e mangle]">(&s, (const <"[$e declarator]"> *)detail))
        return NULL;
      if (bufcat(&s, "}")) return NULL;
      return s;
    }
<'  }'>
    if (bufcat(&s, "{\"ex\":\"%s\"}", <"[$e cname]">_id))
      return NULL;
    return s;
  }
<'}'>

  assert(!"unknown genom exception");
  return NULL;
}




/* --- Exception genom::unkex decoding ------------------------------------- */

static int
json_scan_genom_unkex(genom_unkex_detail *exdetail, const char **json)
{
  const char *e;

  json_skip_whitespace(*json);
  if (*((*json)++) != '{') return EINVAL;
  json_skip_whitespace(*json);
  while(**json != '}') {
    if (*((*json)++) != '"') return EINVAL;

    if (!strncmp(*json, "what\"", sizeof("what"))) {
      (*json) += sizeof("what");
      json_skip_whitespace(*json);
      if (*((*json)++) != ':') return EINVAL;
      json_skip_whitespace(*json);
      if (**json != '"') return EINVAL;
      e = ++(*json);
      while (**json != '"') { if (**json == '\\') (*json)++; (*json)++; }
      if ((size_t)(*json - e) >= sizeof(exdetail->what))
        e = *json - sizeof(exdetail->what);
      strncpy(exdetail->what, e, *json - e);
      exdetail->what[*json - e] = *"";

    } else return EINVAL;

    json_skip_whitespace(*json);
    if (**json == ',') {
      (*json)++;
      json_skip_whitespace(*json);
    }
  }

  (*json)++;
  return 0;
}


/* --- Exception genom::syserr decoding ------------------------------------ */

static int
json_scan_genom_syserr(genom_syserr_detail *exdetail, const char **json)
{
  long long v;
  char *d;

  json_skip_whitespace(*json);
  if (*((*json)++) != '{') return EINVAL;
  json_skip_whitespace(*json);
  while(**json != '}') {
    if (*((*json)++) != '"') return EINVAL;

    if (!strncmp(*json, "code\"", sizeof("code"))) {
      (*json) += sizeof("code");
      json_skip_whitespace(*json);
      if (*((*json)++) != ':') return EINVAL;
      json_skip_whitespace(*json);

      v = strtoll(*json, &d, 0);
      if (d == *json) return EINVAL;
      *json = d;
      exdetail->code = (uint32_t)v;

    } else return EINVAL;

    json_skip_whitespace(*json);
    if (**json == ',') {
      (*json)++;
      json_skip_whitespace(*json);
    }
  }

  (*json)++;
  return 0;
}



/* <"[--- genom_${comp}_decodex -----------------------------------------]"> */

genom_event
genom_<"$comp">_decodex(const char *json,
                void *&exdetail, void (*&exfini)(void *))
{
  const char *p, *e;
  genom_event ex;

  p = json;
  json_skip_whitespace(p);
  if (*(p++) != '{') goto bad;
  json_skip_whitespace(p);
  while(*p != '}') {
    if (*(p++) != '"') goto bad;

    /* look for exception name */
    if (!strncmp(p, "ex\"", sizeof("ex"))) {
      p += sizeof("ex");
      json_skip_whitespace(p);
      if (*(p++) != ':') goto bad;
      json_skip_whitespace(p);
      if (*p != '"') goto bad;
      e = ++p;
      while (*p != '"') { if (*p == '\\') p++; p++; }
<'foreach t [$component throw] {'>
      if (sizeof(<"[$t cname]">_id) == p - e + 1 &&
          !strncmp(<"[$t cname]">_id, e, sizeof(<"[$t cname]">_id)-1))
        ex = <"[$t cname]">_id;
      else
<'}'>
        goto bad;
      p++;

    /* look for exception detail */
    } else if (!strncmp(p, "detail\"", sizeof("detail"))) {
      p += sizeof("detail");
      json_skip_whitespace(p);
      if (*(p++) != ':') goto bad;

<'foreach t [$component throw] {'>
      if (ex == <"[$t cname]">_id) {
<'  if {[llength [$t members]]} {'>
        <"[$t declarator *d]">;
        exdetail = malloc(sizeof(*d));
        if (!exdetail) return genom_fatal;
        d = (<"[$t declarator *]">)exdetail;
        genom::ids::pinit(*d);
        if (json_scan_<"[$t mangle]">(d, &p)) goto bad;
        exfini = genom::ids::pfini< <"[$t declarator]"> >;
        return ex;
<'  } else {'>
        goto bad;
<'  }'>
      }
<'}'>
      if (ex == genom_syserr_id) {
        genom_syserr_detail *d;
        exdetail = malloc(sizeof(*d));
        if (!exdetail) return genom_fatal;
        d = (genom_syserr_detail *)exdetail;
        if (json_scan_genom_syserr(d, &p)) goto bad;
        return ex;
      }
      if (ex == genom_unkex_id) {
        genom_unkex_detail *d;
        exdetail = malloc(sizeof(*d));
        if (!exdetail) return genom_fatal;
        d = (genom_unkex_detail *)exdetail;
        if (json_scan_genom_unkex(d, &p)) goto bad;
        return ex;
      }

    /* unknown json tag */
    } else goto bad;

    json_skip_whitespace(p);
    if (*p == ',') {
      p++;
      json_skip_whitespace(p);
    }
  }

bad:
  genom_unkex_detail *d;
  exdetail = malloc(sizeof(*d));
  if (!exdetail) return genom_fatal;

  d = (genom_unkex_detail *)exdetail;
  strncpy(d->what, json, sizeof(d->what));
  d->what[sizeof(d->what)-1] = *"";
  exfini = NULL;
  return genom_unkex_id;
}
